{% load partials %}
INCLUDED TEMPLATE START
{% partialdef included-partial %}
THIS IS CONTENT FROM THE INCLUDED PARTIAL
{% endpartialdef %}

Now using the partial: {% partial included-partial %}
INCLUDED TEMPLATE END
